Legend:
Page
Library
Module
Module type
Parameter
Class
Class type
Page
Library
Module
Module type
Parameter
Class
Class type
Atd.PredefSourceThe collection of core types known by ATD.
List of all predefined type names with their arity and their definition if applicable.
The parametrized types list, option, etc. all have a definition while the simple types like bool and int have an entry in the list but no associated definition (None).
Create a lookup table from the list of predefined type definitions and extra definitions.
De-alias a type name recursively. This is useful to determine the actual type constructor behind a name. For this, see also get_construct. For example, looking up "special" given the ATD definition type special = int list would return the definition for "list".
Determine the type construct by returning the right-handside of the original type definition associated with type alias. For example, looking up "special" given the ATD definition type special = int list would return (1, List ...) where 1 indicates that the type expression depends on one type parameter.
Determine the type construct associated with a type expression. The result may be Name only if it's a predefined type. If a type name is undefined, the result is None. This is intended for determining default values.