Legend:
Page
Library
Module
Module type
Parameter
Class
Class type
Page
Library
Module
Module type
Parameter
Class
Class type
Atd.PrintSourcePretty-printing of ATD data
Same as format but works on any kind of node wrapped into any e.g. format_any (Variant x).
Turn any AST node into a string. See also format.
Convert a type name with its arguments and its annotations into a string.
Concatenate the components of a type name into a string. For example, TN ["foo"; "bar"] gives "foo.bar". Shorthand for Type_name.to_string.