Legend:
Page
Library
Module
Module type
Parameter
Class
Class type
Page
Library
Module
Module type
Parameter
Class
Class type
Atd.Type_nameSourceUtilities for manipulating type names as found in the ATD AST.
This module can serve as the input to Map or Set functors. e.g.
module Names = Set.Make (Type_name)
The type of a type name.
The simple ATD type name a is represented as TN ["a"]. The composite ATD type name a.b.c is represented as TN ["a"; "b"; "c"].
ATD type definitions normally have a single component. The list of path components may not be empty. Two components indicate a type provided by an external module.
At the moment, the semantics of imports doesn't allow type names of more than two components.
Format to a string in ATD syntax. For example, TN ["a"; "b"] gives a.b.